-- Sourced from Callmore#4947's Scorekeeper, with edits from freeman#2651

rawset(_G, "kvr_autosave" , CV_RegisterVar({
	name = "kvr_autosave",
	defaultvalue = "On",
	flags = CV_NETVAR,
	PossibleValue = CV_OnOff
}))

local FILENAME = "KartVR.txt"
local globalData = {}

-- Here we attempt to open the score file to read the scores into memory.
-- If the file doesent exsist, io.open silently fails.
-- io.open needs to be wrapped inside of an assert to error when a file is
--   failed to be opened.
local f = io.open(FILENAME, "r")
if f then
	-- file already exsists, load from it
	print('Loading scores from "KartVR.txt"...')
	for l in f:lines() do
		local name, kvrscore, kvrwins, kvrlosses, kvrok, kvrrank, kvrrival, kvrrivalwins, kvrrivallosses, kvrrivaldraws, kvrwelcomeback = string.match(l, "(.*);(.*);(.*);(.*);(.*);(.*);(.*);(.*);(.*);(.*);(.*)")

		if name then
			globalData[name] = {tonumber(kvrscore), tonumber(kvrwins), tonumber(kvrlosses), tonumber(kvrok), tonumber(kvrrank), tostring(kvrrival), tonumber(kvrrivalwins), tonumber(kvrrivallosses), tonumber(kvrrivaldraws), tonumber(kvrwelcomeback)}
		end
	end
	f:close()
end

local function _saveFileFunc()
	-- This function actully opens the file.
	-- If the file failed to open for whatever reason, the assert would catch
	--   it and return back to the pcall that called this function.
	local f = assert(io.open(FILENAME, "w"))
	for pn, s in pairs(globalData) do
		if pn:find(";") then continue end -- reject any names that contain a semicolon, since OOPS HOSTS CAN DO THAT LOL
		f:write(pn, ";", s[1], ";", s[2], ";", s[3], ";", s[4], ";", s[5], ";", s[6], ";", s[7], ";", s[8], ";", s[9], ";", s[10], "\n")
	end
	f:close()
end

local function saveFile()
	-- This function is effectivly a wrapper for the real open file function.
	-- The real open file function opens the file through an assert,
	--   if the file failed to open, it would immediatly stop execution and
	--   return here with a false, causing the if block to be ran and to print
	--   the message.
	if consoleplayer ~= server then return end
	--print('Saving data to "KartVR.txt"...')
	if not pcall(_saveFileFunc) then
		print("Failed to save file!")
	end
end

--Because I like the idea of persistent saving.
local function KVR_AutoSave()
	if not kvr_autosave.value then return end
	local pingame = {}
	for p in players.iterate do
		globalData[p.name] = {p.kvrscore, p.kvrwins, p.kvrlosses, p.kvrok, p.kvrrank, p.kvrrival, p.kvrrivalwins, p.kvrrivallosses, p.kvrrivaldraws, p.kvrwelcomeback}
		pingame[p.name] = true
	end
	for n, v in pairs(globalData) do
		if pingame[n] then continue end
	end
	saveFile()
end

local hasRanInt = false

local function think()
	if hasRanInt then
		hasRanInt = false
	end
	
	for p in players.iterate do
		if not p.kvr_loaded then
			p.kvr_loaded = true
			if globalData[p.name] then
				p.kvrscore = globalData[p.name][1]
				p.kvrwins = globalData[p.name][2]
				p.kvrlosses = globalData[p.name][3]
				p.kvrok = globalData[p.name][4]
				p.kvrrank = globalData[p.name][5]
				p.kvrrival = globalData[p.name][6]
				p.kvrrivalwins = globalData[p.name][7]
				p.kvrrivallosses = globalData[p.name][8]
				p.kvrrivaldraws = globalData[p.name][9]
				p.kvrwelcomeback = globalData[p.name][10]
			end
		elseif p.name ~= p.kvr_lastname then
			if globalData[p.name] ~= nil then
				p.kvrscore = globalData[p.name][1]
				p.kvrwins = globalData[p.name][2]
				p.kvrlosses = globalData[p.name][3]
				p.kvrok = globalData[p.name][4]
				p.kvrrank = globalData[p.name][5]
				p.kvrrival = globalData[p.name][6]
				p.kvrrivalwins = globalData[p.name][7]
				p.kvrrivallosses = globalData[p.name][8]
				p.kvrrivaldraws = globalData[p.name][9]
				p.kvrwelcomeback = globalData[p.name][10]
			else
				--check if oldname exsists
				--the and operator here is to stop lower scores from saving over (lol score reset)
				if globalData[p.kvr_lastname] and globalData[p.kvr_lastname][0] then
					if globalData[p.kvr_lastname][0] < p.kvrscore or p.kvrforcesave then
						//force save is here so if someone manually sets the score in-game it still gets saved
						--they exsist, save score and reset
						globalData[p.kvr_lastname][0] = p.kvrscore
						p.kvrforcesave = false
					end
				end
				globalData[p.name] = {1000, 0, 0, 0, 1, "none", 0, 0, 0, 0}
				p.kvrscore = 1000
				p.kvrwins = 0
				p.kvrlosses = 0
				p.kvrok = 0
				p.kvrrank = 1
				p.kvrrival = "none"
				p.kvrrivalwins = 0
				p.kvrrivallosses = 0
				p.kvrrivaldraws = 0
				p.kvrwelcomeback = 0
			end
		end
		p.kvr_lastname = p.name
	end
	
	if leveltime > 236
	and leveltime%(TICRATE*20) == 0 then
		KVR_AutoSave()
	end

end
addHook("ThinkFrame", think)

local autosavetimer = 3*TICRATE
local function voteThink()
	autosavetimer = max($-1, 0)
	
	if autosavetimer == 0
		KVR_AutoSave()
		autosavetimer = 5*TICRATE
	end
end
addHook("VoteThinker", voteThink)

local function intThink()

	if hasRanInt then return end
	hasRanInt = true
	local pingame = {}
	for p in players.iterate do
		globalData[p.name] = {p.kvrscore, p.kvrwins, p.kvrlosses, p.kvrok, p.kvrrank, p.kvrrival, p.kvrrivalwins, p.kvrrivallosses, p.kvrrivaldraws, p.kvrwelcomeback}
		pingame[p.name] = true
	end
	for n, v in pairs(globalData) do
		if pingame[n] then continue end
	end
	saveFile()
end
addHook("IntermissionThinker", intThink)



local function netvars(net)
	globalData = net($)
end
addHook("NetVars", netvars)

local function mapLoad()
	for p in players.iterate do
		if p and p.valid then p.kvr_loaded = false end
	end
end
addHook("MapLoad", mapLoad)