local FRACUNIT = FRACUNIT
local starttime = 236
rawset(_G, "kvr_speedtable", {
	[-1] = "Auto",
	[0] = "Easy",
	[1] = "Normal",
	[2] = "Hard",
	[3] = "Expert"
})

--kvr_kartspeed and kvr_solospeed both control gamespeed, with the main 
--difference being solospeed strictly dictates game speed during free play and 
--doesn't have the Auto option, whereas kartspeed is essentially a substitute 
--for Vanilla's kartspeed command. Set solospeed to your liking if you're running 
--any sort of single-player mods in your server (SPB Attack, TA Leaderboards, 
--etc). kvr_automode determines what rolling Expert does during Auto; "Original" 
--merely sets the game speed to Expert, while "Legacy" will set the game speed 
--to Hard along with also enabling Encore mode on map change - a callback to 
--Mario Kart Wii's netplay experience.

rawset(_G, "kvr_automode" , CV_RegisterVar({
	name = "kvr_automode",
	defaultvalue = "Original",
	flags = CV_NETVAR,
	PossibleValue = {Original = 0, Legacy = 1}
}))

rawset(_G, "kvr_kartspeed" , CV_RegisterVar({
	name = "kvr_kartspeed",
	defaultvalue = "Auto",
	flags = CV_NETVAR|CV_CALL,
	PossibleValue = {Auto = -1, Easy = 0, Normal = 1, Hard = 2, Expert = 3},
	func = function(var)
		if (server) then
			if server.kvrlock then -- Hacky way of replicating hardcode behavior, since it doesn't seem we can properly reference gamestate through lua
				print("\136Game speed will change to \""..kvr_speedtable[var.value].."\" next round.")
			else
				if ((var.value >= 0 and var.value < 2) and var.value ~= gamespeed) then
					COM_BufInsertText(server, "kartspeed "..var.value)
					--print("Kartspeed modified.")
					if server.expertspeed then
						server.expertspeed = false
					end
					if server.autospeed then
						print("\136The host has deactivated Auto Speed.")
						server.autospeed = false
					end
				elseif var.value == 2 then
					if gamespeed ~= 2 then
						COM_BufInsertText(server, "kartspeed hard")
					end
					if server.expertspeed then
						server.expertspeed = false
						print("\136Game Speed has been changed to \"Hard\".")
					end
					
				elseif var.value == 3 then
					if gamespeed ~= 2 then
						COM_BufInsertText(server, "kartspeed hard")
					end
					if not server.expertspeed then
						server.expertspeed = true
						print("\136Game speed has been changed to \"Expert\".")
						--You'll see two prints when switching to Expert from a
						--non-Hard speed setting: One print from hardcode 
						--announcing Vanilla's change to "Hard", and the above 
						--print to clarify we've transitioned to "Expert" speed.
					end
				elseif var.value == -1 then
					if not server.autospeed then
						print("\136Game Speed has been changed to Auto. You and your peers' scores now randomly determine Game Speed.")
						server.autospeed = true
					end
				end
			end
		end
	end
}))

rawset(_G, "kvr_solospeed" , CV_RegisterVar({
	name = "kvr_solospeed",
	defaultvalue = "Expert",
	flags = CV_NETVAR|CV_CALL,
	PossibleValue = {Easy = 0, Normal = 1, Hard = 2, Expert = 3},
	func = function(var)
		if (server) then
			if server.kvrlock then -- Hacky way of replicating hardcode behavior, since it doesn't seem we can properly reference gamestate through lua
				print("\136Game speed will change to \""..kvr_speedtable[var.value].."\" next round.")
			else
				if ((var.value >= 0 and var.value < 2) and var.value ~= gamespeed) then
					COM_BufInsertText(server, "kartspeed "..var.value)
					--print("Kartspeed modified.")
					if server.expertspeed then
						server.expertspeed = false
					end
					if server.autospeed then
						print("\136The host has deactivated Auto Speed.")
						server.autospeed = false
					end
				elseif var.value == 2 then
					if gamespeed ~= 2 then
						COM_BufInsertText(server, "kartspeed hard")
					end
					if server.expertspeed then
						server.expertspeed = false
						print("\136Game Speed has been changed to \"Hard\".")
					end
					
				elseif var.value == 3 then
					if gamespeed ~= 2 then
						COM_BufInsertText(server, "kartspeed hard")
					end
					if not server.expertspeed then
						server.expertspeed = true
						print("\136Game speed has been changed to \"Expert\".")
					end
				end
			end
		end
	end
}))



local SwitchChance
local KVRspeedswitch = false


local ExpertSpeed = FRACUNIT/5 -- This is a surprise tool we'll be using later
local function KVR_ItemSpeed()
	local itemboost = 0
	if server.expertspeed then 
		itemboost = $ + ExpertSpeed
	end
	return itemboost
end

--Giving Items an extra kick during Expert speed
local function kvr_speedUp(mo)
	if not KVR_ItemSpeed() then return end
	--print(GiveItemSpeed()/4)
	local addx = FixedMul(mo.momx, KVR_ItemSpeed()/4)
	local addy = FixedMul(mo.momy, KVR_ItemSpeed()/4)
	P_TryMove(mo, (mo.x + addx), (mo.y + addy), mo.z, false)
end

local itemtypes = {MT_EGGMANITEM, MT_EGGMANITEM_SHIELD, MT_BANANA, MT_BANANA_SHIELD, MT_ORBINAUT, MT_ORBINAUT_SHIELD, MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD,
	MT_SSMINE_SHIELD, MT_SSMINE, MT_BALLHOG, MT_SPB, MT_SINK, MT_SINK_SHIELD}

for mt = 1,#itemtypes do
	addHook("MobjThinker", kvr_speedUp, itemtypes[mt])
end


local function KVR_ExpertMode(p)
	if not server.expertspeed then return end
	local expertboost = ExpertSpeed-(FRACUNIT/20)
	if p.kartstuff[k_bananadrag] < TICRATE and not p.kartstuff[k_offroad] then
		p.kartstuff[k_accelboost] = max($, expertboost/15) --Accel adjust
		p.kartstuff[k_boostpower] = $ + expertboost --Speed adjust
	end
	KVR_ItemSpeed()
end

local lowscore = 0
local lowname = 0
local LobbySize
local avg

--Calculating post-race date for Auto speed to determine potential game speed change
local function KVR_speeddata(p)

	if p.exiting or (p.pflags & PF_TIMEOVER) then
		local KVRsheet = {}
		for d in players.iterate
			if ((d.exiting and not (d.pflags & PF_TIMEOVER)) or (d.pflags & PF_TIMEOVER))
			and d.mo and d.mo.valid then 
				table.insert(KVRsheet, d)
				--print(d.name.." inserted into table")
			end
		end
			
		table.sort(KVRsheet, function(a,b)
			return a.kvrscore < b.kvrscore
		end)
		--print("Table sorted.")

		lowname = KVRsheet[1].name
		lowscore = KVRsheet[1].kvrscore -- The first piece of the puzzle to calculate speed change probabilities
				--print(lowname..": "..lowscore)
		
	end
	
	server.speeddata = true
	LobbySize = KVR_playercount()
	avg = KVR_ratingmean(LobbySize) -- The second piece
	--CONS_Printf(server, "Data collected.")
	--CONS_Printf(server, "Lowest Rating: "..lowname.." - "..lowscore)
	--CONS_Printf(server, "Room Average: "..avg)
	--if IsPlayerAdmin(p)
		--CONS_Printf(p, "Data collected.")
		--CONS_Printf(p, "Lowest Rating: "..lowname.." - "..lowscore)
		--CONS_Printf(p, "Room Average: "..avg)
	--end
end

-- These two functions replicate hardcode's gamespeed parameter on race start
local function KVR_LoadMem()
	if (server) then
		COM_BufInsertText(server, "kvr_kartspeed "..kvr_kartspeed.value)
	end
end

local function KVR_SoloSpeedHandler()
	if (server) then
		if ((kvr_solospeed.value >= 0 and kvr_solospeed.value < 2) and kvr_solospeed.value ~= gamespeed) then
			COM_BufInsertText(server, "kartspeed "..kvr_solospeed.value)
			--print("Kartspeed modified.")
			if server.expertspeed then
				server.expertspeed = false
			end
			if server.autospeed then
				print("\136The host has deactivated Auto Speed.")
				server.autospeed = false
			end
		elseif kvr_solospeed.value == 2 then
			if gamespeed ~= 2 then
				COM_BufInsertText(server, "kartspeed hard")
			end
			if server.expertspeed then
				server.expertspeed = false
			end
			print("\136Game Speed has been changed to \"Hard\".")
		elseif kvr_solospeed.value == 3 then
			if gamespeed ~= 2 then
				COM_BufInsertText(server, "kartspeed hard")
			end
			if not server.expertspeed then
				server.expertspeed = true
				print("\136Game speed has been changed to \"Expert\".")
			end
		end
	end
end

--With how kartencore works in races and voting, the idea behind rolling Encore 
--Hard Speed during Auto is for the encore map to be a temporary thing unless 
--Auto rolls Encore Hard a second time (4% chance). Don't really like the idea 
--of console print spam re: kartencore, nor do I like the idea of normal voting 
--options during kartencore. However, this is hacky at best and will result in a 
--double load - normal load interrupted by Encore taking over. Not entirely 
--obtrusive to me personally, but this certainly requires netgame testing.
addHook("MapChange", function()
	if (server)
		if server.kvrencore
			if not server.encoreload
				COM_BufAddText(server, "map "..G_BuildMapName(gamemap).." -encore") -- Handy Hostmod trick
				server.encoreload = true
			end
		end
	end
end)

addHook("MapLoad", function()
	KVRspeedswitch = true
	server.speeddata = false
	server.votebuffer = 2
	server.voterun = false
	server.kvrlock = false
end)

addHook("ThinkFrame", function()
	if leveltime < 1 then return end
	
	if leveltime == 2 then
		if KVR_ingame() > 1 then
			KVR_LoadMem()
		else
			KVR_SoloSpeedHandler()
		end
		--print(gamestate)
	end
	
	if leveltime == TICRATE
		server.encoreload = false
	end
	if leveltime == 236 then
		if not server.kvrlock then
			server.kvrlock = true
		end
	end

	for p in players.iterate do
		KVR_ExpertMode(p)
		
		if allfinished()
		and kvrcalc
		and not server.speeddata then
			KVR_speeddata(p)
		end
	end
end)

local function KVR_SpeedSelect()
	if KVRspeedswitch then
		KVRspeedswitch = false
		SwitchChance = P_RandomRange(1, 100)
		
		if not server.speeddata then -- Didn't collect post-race data for some reason(RTV, round aborted, etc)
			local KVRsheet = {}
			for p in players.iterate do
				if p and p.valid then
					table.insert(KVRsheet, p)
					--print(p.name.." inserted into table")
				end
			end
			
			table.sort(KVRsheet, function(a,b)
				return a.kvrscore < b.kvrscore
			end)
		
			lowname = KVRsheet[1].name
			lowscore = KVRsheet[1].kvrscore
			print(lowname..": "..lowscore)
			
			LobbySize = KVR_playercount()
			avg = KVR_ratingmean(LobbySize)
		end
		
		--This is where we use lowscore and avg to figure out if Auto changes
		--game speed for the next race. We check lowscore first, then check avg
		--if lowscore is over 500. Percentages and reference values are ported 
		--and modified from CTGP's Friend Room table.
		--(https://wiki.tockdom.com/wiki/Player_Rating)

		if lowscore < 500 then -- Yeah, no advanced speed if score is too low
			if (gamespeed ~= 1 or server.expertspeed or server.kvrencore)
				COM_BufInsertText(server, "kartspeed Normal")
				server.expertspeed = false
				server.kvrencore = false
				chatprint("\136<RatingBot>\137 The next race will be \128Normal Speed\137.", true)
			end
		elseif ((lowscore >= 500 and lowscore < 1000)) then
			if avg < 1200 then
				if (gamespeed ~= 1 or server.expertspeed or server.kvrencore) then
					COM_BufInsertText(server, "kartspeed Normal")
					server.expertspeed = false
					server.kvrencore = false
					chatprint("\136<RatingBot>\137 The next race will be \128Normal Speed\137.", true)
				end
			elseif avg >= 1200 then
				if SwitchChance > 40 then
					if (gamespeed ~= 2 or server.expertspeed or server.kvrencore) then
						COM_BufInsertText(server, "kartspeed Hard")
						server.expertspeed = false
						server.kvrencore = false
						chatprint("\136<RatingBot>\137 The next race will be \133Hard Speed\137.", true)
					end
				else
					if (gamespeed ~= 1 or (server.expertspeed) or (server.kvrencore)) then
						COM_BufInsertText(server, "kartspeed Normal")
						server.expertspeed = false
						server.kvrencore = false
						chatprint("\136<RatingBot>\137 The next race will be \128Normal Speed\137.", true)
					end
				end
			end
		elseif (lowscore >= 1000) then
			if avg < 1700 then
				if SwitchChance > 40 then
					if (gamespeed ~= 2 or (server.expertspeed) or (server.kvrencore)) then
						COM_BufInsertText(server, "kartspeed Hard")
						server.expertspeed = false
						server.kvrencore = false
						chatprint("\136<RatingBot>\137 The next race will be \133Hard Speed\137.", true)
					end
				else
					if (gamespeed ~= 1 or server.expertspeed or server.kvrencore) then
						COM_BufInsertText(server, "kartspeed Normal")
						server.expertspeed = false
						server.kvrencore = false
						chatprint("\136<RatingBot>\137 The next race will be \128Normal Speed\137.", true)
					end
				end
			elseif avg >= 1700 then
				if SwitchChance > 70 then
					if kvr_automode.value then
						if (gamespeed ~=2 or (not server.kvrencore)) then
							COM_BufInsertText(server, "kartspeed Hard")
							server.kvrencore = true
							if server.expertspeed then
								server.expertspeed = false
							end
							chatprint("\136<RatingBot>\137 The next race will be \129Encore Hard Speed\137.", true)
						end
					else
						if (gamespeed ~=2 or (not server.expertspeed)) then
							COM_BufInsertText(server, "kartspeed Hard")
							server.expertspeed = true
							if server.kvrencore then
								server.kvrencore = false
							end
							chatprint("\136<RatingBot>\137 The next race will be \129Expert Speed\137.", true)
						end
					end
				elseif SwitchChance > 20 and SwitchChance <= 70 then
					if (gamespeed ~= 2 or server.expertspeed or server.kvrencore) then
						COM_BufInsertText(server, "kartspeed Hard")
						server.expertspeed = false
						server.kvrencore = false
						chatprint("\136<RatingBot>\137 The next race will be \133Hard Speed\137.", true)
					end
				else
					if (gamespeed ~=1 or server.expertspeed or server.kvrencore) then
						COM_BufInsertText(server, "kartspeed Normal")
						server.expertspeed = false
						server.kvrencore = false
						chatprint("\136<RatingBot>\137 The next race will be \128Normal Speed\137.", true)
					end
				end
			end				
		end
	end
end

addHook("VoteThinker", function()
	server.votebuffer = max($-1, 0) --Because tic 0 is fucking cursed
	if server.votebuffer == 0
	and not server.voterun then
		if KVR_ingame() > 1 then
			if kvr_kartspeed.value == -1 then
				KVR_SpeedSelect()
			else
				KVR_LoadMem()
			end
		else
			if server.kvrencore then
				server.kvrencore = false
			end
		end
		server.voterun = true
	end
end)

addHook("IntermissionThinker", function()
	server.kvrlock = false
	if CV_FindVar("advancemap").string:lower() ~= "vote" then
		KVR_SpeedSelect()
	end
end)

addHook("NetVars", function(net)
	SwitchChance = net($)
	KVRspeedswitch = net($)
	lowscore = net($)
	lowname = net($)
end)